
/* Generated data (by glib-mkenums) */

#ifndef __GDU_ENUM_TYPES_H__
#define __GDU_ENUM_TYPES_H__

#include <glib-object.h>

G_BEGIN_DECLS

/* enumerations from "gduenums.h" */
GType gdu_volume_grid_element_type_get_type (void) G_GNUC_CONST;
#define GDU_TYPE_VOLUME_GRID_ELEMENT_TYPE (gdu_volume_grid_element_type_get_type ())
GType gdu_power_state_flags_get_type (void) G_GNUC_CONST;
#define GDU_TYPE_POWER_STATE_FLAGS (gdu_power_state_flags_get_type ())
GType gdu_device_tree_model_flags_get_type (void) G_GNUC_CONST;
#define GDU_TYPE_DEVICE_TREE_MODEL_FLAGS (gdu_device_tree_model_flags_get_type ())
G_END_DECLS

#endif /* __GDU_ENUM_TYPES_H__ */

/* Generated data ends here */

