
/* Generated data (by glib-mkenums) */

#ifndef __LIB_GDU_ENUM_TYPES_H__
#define __LIB_GDU_ENUM_TYPES_H__

#include <glib-object.h>

G_BEGIN_DECLS

/* enumerations from "libgduenums.h" */
GType gdu_format_duration_flags_get_type (void) G_GNUC_CONST;
#define GDU_TYPE_FORMAT_DURATION_FLAGS (gdu_format_duration_flags_get_type ())
G_END_DECLS

#endif /* __LIB_GDU_ENUM_TYPES_H__ */

/* Generated data ends here */

